unit EvalExpr;

interface

uses SysUtils, Classes, Variants, ScriptParser, TreeList, ScriptVariants,
  ScriptFuncs, Interprt;

type
  EExpressionError = class(Exception);

type
  TEvaluateExpr = class(TObject)
  private
    FParser: TCustomParser;
    //FExpression: string;
    FInterpreter: TInterpreter;
    FObjectResult: TObject;
    FResultType: Integer;
  protected
    function EvaluateNode(P: TNodeList): Variant;
  public
    constructor Create(AInterpreter: TInterpreter; Stream: TStream);
    destructor Destroy; override;
    function Evaluate: Variant;
    procedure Error(const Ident: string);
    procedure ExpressionError(const Message: string);
    procedure ErrorStr(const Message: string);
    procedure ErrorFmt(const Ident: string; Args: array of const);
    property Parser: TCustomParser read FParser;
    property Interpreter: TInterpreter read FInterpreter;
    property ObjectResult: TObject read FObjectResult write FObjectResult;
    property ResultType: Integer read FResultType write FResultType;
  end;


implementation

const
  SExpressionError = '[Expression Error] %s (%d)';
  SExpressionExpected = 'Expression expected.';
  
var
  Level: Integer = 0;

procedure PrintNodes(P: TNodeList);
var
  N: TNodeItem;
  C: TNodeList;
begin
  if Assigned(P) then begin
    N := P.First;
    while N <> nil do begin
      WriteLn(StringOfChar(#9, Level), N.Value);
      if N.HasChildren then begin
        Inc(Level);
        C := N.FirstChild;
        while C <> nil do begin
          PrintNodes(C);
          C := C.Next;
        end;
        Dec(Level);
      end;
      N := N.Next;
    end;
  end;
end;

{ TEvaluateExpr }

constructor TEvaluateExpr.Create(AInterpreter: TInterpreter; Stream: TStream);
begin
  inherited Create;
  FInterpreter := AInterpreter;
  FParser :=  TCustomParser.Create(Stream);
end;

destructor TEvaluateExpr.Destroy;
begin
  FParser.Free;
  inherited Destroy;
end;

procedure TEvaluateExpr.Error(const Ident: string);
begin
  ErrorStr(Ident);
end;

procedure TEvaluateExpr.ErrorFmt(const Ident: string;
  Args: array of const);
begin
  ErrorStr(Format(Ident, Args));
end;

procedure TEvaluateExpr.ErrorStr(const Message: string);
begin
  raise EExpressionError.Create(Message);
end;

function TEvaluateExpr.Evaluate: Variant;
var
  Root: TNodeList;

  { PrepareTreeList }
  procedure PrepareTreeList;
  var
    N: TNodeItem;
    Token: Word;
    C: TNodeList;
  begin
    C := Root;
    while True do begin
      Token := Parser.NextToken;
      case Token of

        idEOF: Break;

        idInteger:
          begin
            N := C.Append(Parser.TokenInt);
            N.NodeType := idInteger;
          end;

        idFloat:
          begin
            N := C.Append(Parser.TokenFloat);
            N.NodeType := idFloat;
          end;

        idLParenthes:
          begin
            N := C.Append('()');
            N.NodeType := idLParenthes;
            C := N.AppendChild;
          end;

        idRParenthes:
          begin
            C := C.Parent.Owner;
          end;

        idComma:
          begin
            C := C.Parent.AppendChild;
          end;
          
        idString:
          begin
            N := C.Append(Parser.TokenString);
            N.NodeType := idString;
          end;

        idIdent:
          begin
            N := C.Append(Parser.TokenString);
            N.NodeType := idIdent;
          end;

        idSlash,
        idLessThan,
        idGreaterThan,
        idEqual,
        idNotEqual,
        idAssignment,
        idSemicolon,
        idPlus,
        idMinus,
        idMul,
        idDiv,
        idMod,
        idLessEqual,
        idGreaterEqual,
        idTilde,
        idExclamation,
        idNot,
        idAnd,
        idShr,
        idShl,
        idOr,
        idXor:
          begin
            N := C.Append(Parser.TokenString);
            N.NodeType := Token;
          end;

      end;
    end;
  end;

begin
  Result := Unassigned;
  Root := TNodeList.Create(nil);
  try
    try
      PrepareTreeList;
      //PrintNodes(Root);
      Result := EvaluateNode(Root);
    except
      on E: Exception do ExpressionError(E.Message);
    end;
  finally
    Root.Free;
  end;
end;

function TEvaluateExpr.EvaluateNode(P: TNodeList): Variant;
var
  N: TNodeItem;
  C: TNodeList;

  procedure ScanVariants(L: TNodeList);
  var
    Item: TNodeItem;
    VarItem: TScriptVariant;
    Ident: string;
  begin
    if Assigned(L) then begin
      Item := L.First;
      while Item <> nil do begin
        if Item.NodeType = idIdent then begin
          Ident := Item.Value;
          if ScriptVariantList.IsDefined(Ident) then
          begin
            VarItem := ScriptVariantList.GetVariant(Ident);
            Item.Value := Unassigned;
            Item.ObjectValue := VarItem.ObjectValue;
            Item.Value := VarItem.Value;
            Item.NodeType := VarItem.VarType;
          end
          else if ScriptConstantList.IsDefined(Ident) then
          begin
            VarItem := ScriptConstantList.GetVariant(Ident);
            Item.Value := Unassigned;
            Item.ObjectValue := VarItem.ObjectValue;
            Item.Value := VarItem.Value;
            Item.NodeType := VarItem.VarType;
          end
          else if not ScriptFunctionList.IsDefined(Ident) then
            ErrorFmt( 'Unknown identifier: "%s"', [Ident] );
        end;
        Item := Item.Next;
      end;
    end;
  end;

  procedure ScanFunctions(L: TNodeList);
  var
    Item: TNodeItem;
    Child: TNodeList;
    Func: TScriptFunction;
    Param: TScriptParam;
    Count: Integer;
    Result: Variant;
    Ident: string;
    ParamCount: Integer;
  begin
    if Assigned(L) then
    begin
      Item := L.First;
      while Item <> nil do
      begin
        if Item.NodeType = idIdent then
        begin
          Ident := Item.Value;
          if ScriptFunctionList.IsDefined(Ident) then
          begin
            Child := Item.Next.FirstChild;
            Func := ScriptFunctionList.Functions[Ident];

            ParamCount := Func.ParamCount;
            if Func.ParamCount = -1 then // OpenParam
              Func.Params.Clear;

            Count := 0;
            while Child <> nil do
            begin
              // Buraya dikkat! Paramatersiz fonksiyonlar icin!
              if not Assigned( Child.First ) then Break;

              if ParamCount = -1 then // OpenParam
              begin
                Param := Func.Params.AddParam;
                with Param do begin
                  Name := '@Arg' + IntToStr(Count);
                  Value := Child.First.Value;
                  ObjectValue := Child.First.ObjectValue;
                  ParamType := Child.First.NodeType;
                end;
              end
              else begin
                Param := Func.Params[Count];
                Param.Value := Child.First.Value;
                Param.ObjectValue := Child.First.ObjectValue;
                Param.ParamType := Child.First.NodeType;
                if Param.Options = idVariant then begin
                
                end;
              end;
              Inc(Count);
              Child := Child.Next;
            end;
            
            Result := Func.Execute;

            if Func.ParamCount = -1 then // OpenParam
              Func.Params.Clear;
              
            L.Delete(Item.Next);
            Item.Value := Unassigned;
            Item.Value := Result;
            Item.ObjectValue := Func.ObjectResult;
            Item.NodeType := Func.ResultType;
          end
          else ErrorFmt( 'Unknown identifier: "%s"', [Ident] );
        end;
        Item := Item.Next;
      end;
    end;
  end;

  function ScanExpression(L: TNodeList): Variant;
  var
    Item, Left, Right: TNodeItem;
    Parent: TNodeItem;
  begin
    Result := Unassigned;
    if Assigned(L) then begin
    
      // not
      Item := L.First;
      while Item <> nil do
      begin
        if Item.NodeType = idNot then
        begin
          Right := Item.Next;
          if not Assigned( Right ) then
            Error(SExpressionExpected);
          Right.Value := not Right.Value;
          L.Delete(Item);
          Item := Right;
        end;
        Item := Item.Next;
      end;

      // *, /, div, mod, and, shr, shl
      Item := L.First;
      while Item <> nil do
      begin

        if (Item.NodeType = idMul) or (Item.NodeType = idSlash) or
           (Item.NodeType = idDiv) or (Item.NodeType = idMod) or
           (Item.NodeType = idAnd) or (Item.NodeType = idShl) or
           (Item.NodeType = idShr) then
        begin

          Left := Item.Prev;
          if not Assigned( Left ) then
            Error(SExpressionExpected);
          Right := Item.Next;
          if not Assigned( Right ) then
            Error(SExpressionExpected);

          case Item.NodeType of
            idMul:   Left.Value := Left.Value * Right.Value;
            idSlash: Left.Value := Left.Value / Right.Value;
            idDiv:   Left.Value := Left.Value div Right.Value;
            idMod:   Left.Value := Left.Value mod Right.Value;
            idAnd:   Left.Value := Left.Value and Right.Value;
            idShl:   Left.Value := Left.Value shl Right.Value;
            idShr:   Left.Value := Left.Value shr Right.Value;
          end;

          L.Delete(Right);
          L.Delete(Item);
          Item := Left;
          
        end;
        {
        if Item.NodeType = idMul then
        begin
          Left := Item.Prev;
          if not Assigned( Left ) then
            Error(SExpressionExpected);
          Right := Item.Next;
          if not Assigned( Right ) then
            Error(SExpressionExpected);
          Left.Value := Left.Value * Right.Value;
          L.Delete(Right);
          L.Delete(Item);
          Item := Left;
        end
        else if Item.NodeType = idSlash then
        begin
          Left := Item.Prev;
          if not Assigned( Left ) then
            Error(SExpressionExpected);
          Right := Item.Next;
          if not Assigned( Right ) then
            Error(SExpressionExpected);
          Left.Value := Left.Value / Right.Value;
          L.Delete(Right);
          L.Delete(Item);
          Item := Left;
        end
        else if Item.NodeType = idDiv then
        begin
          Left := Item.Prev;
          if not Assigned( Left ) then
            Error(SExpressionExpected);
          Right := Item.Next;
          if not Assigned( Right ) then
            Error(SExpressionExpected);
          Left.Value := Left.Value div Right.Value;
          L.Delete(Right);
          L.Delete(Item);
          Item := Left;
        end
        else if Item.NodeType = idMod then
        begin
          Left := Item.Prev;
          if not Assigned( Left ) then
            Error(SExpressionExpected);
          Right := Item.Next;
          if not Assigned( Right ) then
            Error(SExpressionExpected);
          Left.Value := Left.Value mod Right.Value;
          L.Delete(Right);
          L.Delete(Item);
          Item := Left;
        end
        else if Item.NodeType = idAnd then
        begin
          Left := Item.Prev;
          if not Assigned( Left ) then
            Error(SExpressionExpected);
          Right := Item.Next;
          if not Assigned( Right ) then
            Error(SExpressionExpected);
          Left.Value := Left.Value and Right.Value;
          L.Delete(Right);
          L.Delete(Item);
          Item := Left;
        end
        else if Item.NodeType = idShl then
        begin
          Left := Item.Prev;
          if not Assigned( Left ) then
            Error(SExpressionExpected);
          Right := Item.Next;
          if not Assigned( Right ) then
            Error(SExpressionExpected);
          Left.Value := Left.Value shl Right.Value;
          L.Delete(Right);
          L.Delete(Item);
          Item := Left;
        end
        else if Item.NodeType = idShr then
        begin
          Left := Item.Prev;
          if not Assigned( Left ) then
            Error(SExpressionExpected);
          Right := Item.Next;
          if not Assigned( Right ) then
            Error(SExpressionExpected);
          Left.Value := Left.Value shr Right.Value;
          L.Delete(Right);
          L.Delete(Item);
          Item := Left;
        end;
        }
        
        Item := Item.Next;
      end;


      // Plus, Minus, Or, Xor
      Item := L.First;
      while Item <> nil do
      begin
        if Item.NodeType = idPlus then
        begin
          Left := Item.Prev;
          if Left = nil then
          begin
            L.Delete(Item);
            Item := L.First;
            Continue;
          end
          else if (Left.NodeType = idEqual) or
             (Left.NodeType = idNotEqual) or
             (Left.NodeType = idLessEqual) or
             (Left.NodeType = idGreaterEqual) or
             (Left.NodeType = idLessThan) or
             (Left.NodeType = idGreaterThan) then begin
            L.Delete(Item);
            Item := L.First;
            Continue;
          end;
          Right := Item.Next;
          if not Assigned( Right ) then
            Error(SExpressionExpected);
          Left.Value := Left.Value + Right.Value;
          L.Delete(Right);
          L.Delete(Item);
          Item := Left;
        end
        else
        if Item.NodeType = idMinus then
        begin
          Left := Item.Prev;
          if Left = nil then
          begin
            Right := Item.Next;
            if not Assigned( Right ) then
              Error(SExpressionExpected);
            Right.Value := - Right.Value;
            L.Delete(Item);
            Item := L.First;
            Continue;
          end
          else if (Left.NodeType = idEqual) or
             (Left.NodeType = idNotEqual) or
             (Left.NodeType = idLessEqual) or
             (Left.NodeType = idGreaterEqual) or
             (Left.NodeType = idLessThan) or
             (Left.NodeType = idGreaterThan) then begin
            Right := Item.Next;
            if not Assigned( Right ) then
              Error(SExpressionExpected);
            Right.Value := - Right.Value;
            L.Delete(Item);
            Item := L.First;
            Continue;
          end;
          Right := Item.Next;
          if not Assigned( Right ) then
            Error(SExpressionExpected);
          Left.Value := Left.Value - Right.Value;
          L.Delete(Right);
          L.Delete(Item);
          Item := Left;
        end
        else
        if Item.NodeType = idOr then
        begin
          Left := Item.Prev;
          if not Assigned( Left ) then
            Error(SExpressionExpected);
          Right := Item.Next;
          if not Assigned( Right ) then
            Error(SExpressionExpected);
          Left.Value := Left.Value or Right.Value;
          L.Delete(Right);
          L.Delete(Item);
          Item := Left;
        end
        else if Item.NodeType = idXor then
        begin
          Left := Item.Prev;
          if not Assigned( Left ) then
            Error(SExpressionExpected);
          Right := Item.Next;
          if not Assigned( Right ) then
            Error(SExpressionExpected);
          Left.Value := Left.Value xor Right.Value;
          L.Delete(Right);
          L.Delete(Item);
          Item := Left;
        end;
        Item := Item.Next;
      end;

      // Equal, not Equal, Less, Greater, LessEqual, GreaterEqual
      Item := L.First;
      while Item <> nil do
      begin

        if (Item.NodeType = idEqual) or (Item.NodeType = idNotEqual) or
           (Item.NodeType = idLessThan) or (Item.NodeType = idGreaterThan) or
           (Item.NodeType = idLessEqual) or (Item.NodeType = idGreaterEqual) then
        begin

          Left := Item.Prev;
          if not Assigned( Left ) then
            Error(SExpressionExpected);
          Right := Item.Next;
          if not Assigned( Right ) then
            Error(SExpressionExpected);

          case Item.NodeType of
            idEqual:        Left.Value := Left.Value = Right.Value;
            idNotEqual:     Left.Value := Left.Value <> Right.Value;
            idLessThan:     Left.Value := Left.Value < Right.Value;
            idGreaterThan:  Left.Value := Left.Value > Right.Value;
            idLessEqual:    Left.Value := Left.Value <= Right.Value;
            idGreaterEqual: Left.Value := Left.Value >= Right.Value;
          end;

          L.Delete(Right);
          L.Delete(Item);
          Item := Left;
          
        end;
        {
        if Item.NodeType = idEqual then begin
          Left := Item.Prev;
          if not Assigned( Left ) then
            Error(SExpressionExpected);
          Right := Item.Next;
          if not Assigned( Right ) then
            Error(SExpressionExpected);
          Left.Value := Left.Value = Right.Value;
          L.Delete(Right);
          L.Delete(Item);
          Item := Left;
        end
        else
        if Item.NodeType = idNotEqual then begin
          Left := Item.Prev;
          if not Assigned( Left ) then
            Error(SExpressionExpected);
          Right := Item.Next;
          if not Assigned( Right ) then
            Error(SExpressionExpected);
          Left.Value := Left.Value <> Right.Value;
          L.Delete(Right);
          L.Delete(Item);
          Item := Left;
        end
        else
        if Item.NodeType = idLessThan then begin
          Left := Item.Prev;
          if not Assigned( Left ) then
            Error(SExpressionExpected);
          Right := Item.Next;
          if not Assigned( Right ) then
            Error(SExpressionExpected);
          Left.Value := Left.Value < Right.Value;
          L.Delete(Right);
          L.Delete(Item);
          Item := Left;
        end
        else
        if Item.NodeType = idGreaterThan then begin
          Left := Item.Prev;
          if not Assigned( Left ) then
            Error(SExpressionExpected);
          Right := Item.Next;
          if not Assigned( Right ) then
            Error(SExpressionExpected);
          Left.Value := Left.Value > Right.Value;
          L.Delete(Right);
          L.Delete(Item);
          Item := Left;
        end
        else
        if Item.NodeType = idLessEqual then begin
          Left := Item.Prev;
          if not Assigned( Left ) then
            Error(SExpressionExpected);
          Right := Item.Next;
          if not Assigned( Right ) then
            Error(SExpressionExpected);
          Left.Value := Left.Value <= Right.Value;
          L.Delete(Right);
          L.Delete(Item);
          Item := Left;
        end
        else
        if Item.NodeType = idGreaterEqual then begin
          Left := Item.Prev;
          if not Assigned( Left ) then
            Error(SExpressionExpected);
          Right := Item.Next;
          if not Assigned( Right ) then
            Error(SExpressionExpected);
          Left.Value := Left.Value >= Right.Value;
          L.Delete(Right);
          L.Delete(Item);
          Item := Left;
        end;
        }
        Item := Item.Next;
      end;

      // Problemli satirlar. Mutlaka duzenlenmeli
      Item := L.First;
      if Assigned(Item) then
      begin
        Result := Item.Value;
        ObjectResult := Item.ObjectValue;
        ResultType := Item.NodeType;
        if Assigned(Item.Owner.Parent) then
        begin
          Parent := Item.Owner.Parent;
          if Parent.NodeType = idLParenthes then
          begin
            if Assigned(Parent.Prev) then
            begin
              if Parent.Prev.NodeType <> idIdent then
              begin
                Parent.Value := Unassigned;
                Parent.NodeType := Item.NodeType;
                Parent.Value := Item.Value;
                Parent.ObjectValue := Item.ObjectValue;
                //ObjectResult := Item.ObjectValue;
                //Parent.DeleteChild(Item.Owner);
              end;
            end
            else begin
              Parent.Value := Unassigned;
              Parent.NodeType := Item.NodeType;
              Parent.Value := Item.Value;
              Parent.ObjectValue := Item.ObjectValue;
              //ObjectResult := Item.ObjectValue;
              //Parent.DeleteChild(Item.Owner);
            end;
          end;
        end;
      end
      else begin
        Parent := L.Parent;
        if Assigned(Parent) then
        begin
          if Parent.NodeType = idLParenthes then
          begin
            Parent.Value := Unassigned;
            Parent.NodeType := idUnknown;
            Parent.ObjectValue := nil;
          end;
        end;
      end;
      // Problemli satirlar. Mutlaka duzenlenmeli - Son

    end;

  end;

  function EvaluateList(L: TNodeList): Variant;
  begin
    {
      1. Variants
      2. Functions
      3. Not, Mul, Div, Plus, Minus, ...
    }
    Result := Unassigned;
    ScanVariants(L);
    ScanFunctions(L);
    Result := ScanExpression(L);
  end;
  
begin
  Result := Unassigned;
  if Assigned(P) then begin
    N := P.First;
    while N <> nil do begin
      if N.HasChildren then begin
        C := N.FirstChild;
        while C <> nil do begin
          EvaluateNode(C);
          C := C.Next;
        end;
      end;
      N := N.Next;
    end;
    Result := EvaluateList(P);
  end;
end;

procedure TEvaluateExpr.ExpressionError(const Message: string);
begin
  if Assigned(Interpreter) then
    Interpreter.Error(Message)
  else
    raise EExpressionError.Create('<b>' + Message + '</b>');
end;

end.
