unit Execution;

interface

uses SysUtils, Classes, Contnrs, Math, Variants, ScriptParser, ScriptVariants,
  ScriptFuncs;

const
  ecNop       = 0;
  ecHtml      = 1;
  ecAssign    = 2;
  ecCall      = 3;
  ecWhile     = 4;
  ecEndWhile  = 5;
  ecRepeat    = 6;
  ecUntil     = 7;
  ecIf        = 8;
  ecElse      = 9;
  ecEndIf     = 10;
  ecJump      = 11;
  ecBreak     = 12;
  ecContinue  = 13;

type
  { TExecutionItem }
  TExecutionItem = class(TObject)
  private
    FCode: Integer;
    FIdent: string; // for Assign
    FText: string; // pure text
    FJumpCode: Integer;
  public
    constructor Create;
    destructor Destroy; override;
    property Code: Integer read FCode write FCode;
    property JumpCode: Integer read FJumpCode write FJumpCode;
    property Ident: string read FIdent write FIdent;
    property Text: string read FText write FText;
  end;

  { TExecutionList }
  TExecutionList = class(TObjectList)
  protected
    function GetItem(Index: Integer): TExecutionItem;
    procedure SetItem(Index: Integer; AItem: TExecutionItem);
  public
    function Add(AItem: TExecutionItem): Integer;
    function Remove(AItem: TExecutionItem): Integer;
    function IndexOf(AItem: TExecutionItem): Integer;
    procedure Insert(Index: Integer; AItem: TExecutionItem);
    function AddHtml(const Html: string): Integer;
    function AddAssign(const IdentDef, PureText: string): Integer;
    function AddCall(const IdentDef, PureText: string): Integer;
    function AddJump: Integer;
    function AddWhile(const PureText: string): Integer;
    function AddRepeat: Integer;
    function AddUntil(const PureText: string): Integer;
    function AddIf(const PureText: string): Integer;
    function AddElse: Integer;
    function AddBreak: Integer;
    function AddContinue: Integer;
    property Items[Index: Integer]: TExecutionItem read GetItem write SetItem; default;
  end;

var
  ExecutionList: TExecutionList;

implementation

{ InitExecutionList }
procedure InitExecutionList;
begin
  ExecutionList := TExecutionList.Create;
end;

{ DoneExecutionList }
procedure DoneExecutionList;
begin
  ExecutionList.Free;
end;

{ TExecutionItem }

constructor TExecutionItem.Create;
begin
  inherited Create;
end;

destructor TExecutionItem.Destroy;
begin
  inherited Destroy;
end;

{ TExecutionList }

function TExecutionList.Add(AItem: TExecutionItem): Integer;
begin
  Result := inherited Add(AItem);
end;

function TExecutionList.AddAssign(const IdentDef, PureText: string): Integer;
var
  Item: TExecutionItem;
begin
  Item := TExecutionItem.Create;
  with Item do begin
    Code := ecAssign;
    Ident := IdentDef;
    Text := PureText;
  end;
  Result := ExecutionList.Add(Item);
end;

function TExecutionList.AddBreak: Integer;
var
  Item: TExecutionItem;
begin
  Item := TExecutionItem.Create;
  with Item do
    Code := ecBreak;
  Result := ExecutionList.Add(Item);
end;

function TExecutionList.AddCall(const IdentDef, PureText: string): Integer;
var
  Item: TExecutionItem;
begin
  Item := TExecutionItem.Create;
  with Item do begin
    Code := ecCall;
    Ident := IdentDef;
    Text := PureText;
  end;
  Result := ExecutionList.Add(Item);
end;

function TExecutionList.AddContinue: Integer;
var
  Item: TExecutionItem;
begin
  Item := TExecutionItem.Create;
  with Item do
    Code := ecContinue;
  Result := ExecutionList.Add(Item);
end;

function TExecutionList.AddElse: Integer;
var
  Item: TExecutionItem;
begin
  Item := TExecutionItem.Create;
  with Item do
    Code := ecElse;
  Result := ExecutionList.Add(Item);
end;

function TExecutionList.AddHtml(const Html: string): Integer;
var
  Item: TExecutionItem;
begin
  Item := TExecutionItem.Create;
  with Item do begin
    Code := ecHtml;
    Text := Html;
  end;
  Result := ExecutionList.Add(Item);
end;

function TExecutionList.AddIf(const PureText: string): Integer;
var
  Item: TExecutionItem;
begin
  Item := TExecutionItem.Create;
  with Item do begin
    Code := ecIf;
    Text := PureText;
  end;
  Result := ExecutionList.Add(Item);
end;

function TExecutionList.AddJump: Integer;
var
  Item: TExecutionItem;
begin
  Item := TExecutionItem.Create;
  with Item do
    Code := ecJump;
  Result := ExecutionList.Add(Item);
end;

function TExecutionList.AddRepeat: Integer;
var
  Item: TExecutionItem;
begin
  Item := TExecutionItem.Create;
  with Item do
    Code := ecRepeat;
  Result := ExecutionList.Add(Item);
end;

function TExecutionList.AddUntil(const PureText: string): Integer;
var
  Item: TExecutionItem;
begin
  Item := TExecutionItem.Create;
  with Item do begin
    Code := ecUntil;
    Text := PureText;
  end;
  Result := ExecutionList.Add(Item);
end;

function TExecutionList.AddWhile(const PureText: string): Integer;
var
  Item: TExecutionItem;
begin
  Item := TExecutionItem.Create;
  with Item do begin
    Code := ecWhile;
    Text := PureText;
  end;
  Result := ExecutionList.Add(Item);
end;

function TExecutionList.GetItem(Index: Integer): TExecutionItem;
begin
  Result := TExecutionItem(inherited Items[Index]);
end;

function TExecutionList.IndexOf(AItem: TExecutionItem): Integer;
begin
  Result := inherited IndexOf(AItem);
end;

procedure TExecutionList.Insert(Index: Integer; AItem: TExecutionItem);
begin
  inherited Insert(Index, AItem);
end;

function TExecutionList.Remove(AItem: TExecutionItem): Integer;
begin
  Result := inherited Remove(AItem);
end;

procedure TExecutionList.SetItem(Index: Integer; AItem: TExecutionItem);
begin
  inherited Items[Index] := AItem;
end;

initialization
  InitExecutionList;

finalization
  DoneExecutionList;

end.
