
{.$DEFINE DEBUG}

unit Interprt;

interface

uses SysUtils, Classes, Contnrs, Math, Variants, ScriptParser, ScriptVariants,
  ScriptFuncs, Execution;

type
  EInterpreterError = class(Exception);

type
  TInterpreter = class(TObject)
  private
    FScriptName: string;
    FParser: TCustomParser;
    function TakePureText(var CurrToken: Word; const ToToken: Word): string;
  private
    procedure CompileIdent;
    procedure CompileIncludeBlock;
    procedure CompileVarBlock;
    procedure CompileConstBlock;
    procedure CompileAssignBlock(const Ident: string);
    procedure CompileFunctionBlock(const Ident: string);
    procedure CompileWhileBlock;
    procedure CompileRepeatBlock;
    procedure CompileHtmlBlock;
    procedure CompileIfBlock;
    procedure CompileEvaluateBlock;
    procedure CompileBlock;
    procedure CompileBreak;
    procedure CompileContinue;
    procedure CompileForBlock;
    procedure RunAssign(Item: TExecutionItem);
    procedure RunCall(Item: TExecutionItem);
    procedure RunWhile(Item: TExecutionItem; var Index: Integer);
    procedure RunRepeat(Item: TExecutionItem; var Index: Integer);
    procedure RunUntil(Item: TExecutionItem; var Index: Integer);
    procedure RunJump(Item: TExecutionItem; var Index: Integer);
    procedure RunIf(Item: TExecutionItem; var Index: Integer);
    procedure RunElse(Item: TExecutionItem; var Index: Integer);
    procedure RunEndIf(Item: TExecutionItem; var Index: Integer);
    procedure RunBreak(Item: TExecutionItem; var Index: Integer);
    procedure RunContinue(Item: TExecutionItem; var Index: Integer);
  public
    constructor Create;
    destructor Destroy; override;
    procedure CompileFile(const FileName: string);
    procedure Error(const Ident: string);
    procedure ErrorStr(const Message: string);
    procedure ErrorFmt(const Ident: string; Args: array of const);
    procedure Run;
    function NextToken: Word;
    property ScriptName: string read FScriptName;
    property Parser: TCustomParser read FParser;
  end;

const
  SInterpreterError = '[Error] %s(%d): %s';

implementation

uses CanApp, EvalExpr;

{ TInterpreter }

procedure TInterpreter.CompileAssignBlock(const Ident: string);
var
  DefText: string;
  Token: Word;
begin
  {
  if not ScriptVariantList.IsDefined(Ident) then
    ErrorFmt( '"%s" not defined.', [Ident] );

  Token := NextToken;

  if Token = idAssigment then
  }
  
  if ScriptVariantList.IsDefined(Ident) then
  begin
    Token := NextToken;
    if Token = idAssignment then
    begin
      DefText := TakePureText( Token, idSemicolon );
      {
      Token := NextToken;
      while Token <> idSemicolon do
      begin
        if (Token = idEOF) or (Token = idElse) then Break;
        if Token = idString then
          DefText := DefText + GetQuotedStr(Parser.TokenString)
        else DefText := DefText + ' ' + Parser.TokenString;
        Token := NextToken;
      end;
      }
      if DefText <> '' then
        ExecutionList.AddAssign(Ident, DefText)
      else Error('Expression expected.');
    end
    else Error( '":=" symbol expected.' );
  end
  else ErrorFmt( '"%s" not defined.', [Ident] );
end;

procedure TInterpreter.CompileBlock;
begin
  case Parser.Token of
    idVar       : CompileVarBlock;
    idConst     : CompileConstBlock;
    idInclude   : CompileIncludeBlock;
    idUses      : CompileIncludeBlock; // Same as Include
    idWhile     : CompileWhileBlock;
    idRepeat    : CompileRepeatBlock;
    idIf        : CompileIfBlock;
    idIdent     : CompileIdent;
    idHtml      : CompileHtmlBlock;
    idEvaluate  : CompileEvaluateBlock;
    idComment   :;
    idScriptEnd :;
    idScriptBegin:;
    idNullHtml:;
    idSemicolon:; // Duzeltmeyi Unutma!
    idBegin     :;
    idBreak     : CompileBreak;
    idContinue  : CompileContinue;
    idFor       : CompileForBlock;
  else
    ErrorFmt( 'Unknown identifier: "%s"', [Parser.TokenString] );
  end;
end;

procedure TInterpreter.CompileBreak;
begin
  ExecutionList.AddBreak;
  NextToken; // skip "break" keyword
end;

procedure TInterpreter.CompileConstBlock;
var
  Ident, DefText: string;
  Token: Word;
  Value: Variant;

  EvaluateExpr: TEvaluateExpr;
  Stream: TStringStream;

begin
  Token := NextToken;
  if Token = idIdent then
  begin
    Ident := Parser.TokenString;
    if IdentisKeyword(Ident) then
      ErrorFmt( 'Constant is keyword: "%s"', [Ident] );
    if not ScriptConstantList.IsDefined(Ident) then
    begin
      Token := NextToken;
      if Token = idEqual then begin
        DefText := TakePureText( Token, idSemicolon );
        if DefText <> '' then begin
          Stream := TStringStream.Create('<% ' + DefText + ' %>' );
          try
            EvaluateExpr := TEvaluateExpr.Create(Self, Stream);
            try
              Value := EvaluateExpr.Evaluate;
              {
              ScriptConstantList.Add(TScriptVariant.Create(Ident,
                Value, VarTypeToId(VarType(Value))));
              }
              ScriptConstantList.CreateVar(Ident, Value, VarTypeToId(VarType(Value)));
              {
              ScriptVar := ScriptConstantList.AddVar;
              ScriptVar.Name := Ident;
              ScriptVar.Value := Value;
              ScriptVar.VarType := VarTypeToId(VarType(Value));
              }
            finally
              EvaluateExpr.Free;
            end;
          finally
            Stream.Free;
          end;
        end
        else Error('Expression expected.');
      end
      else Error('"=" symbol expected.');
    end
    else ErrorFmt( 'Constant already defined: "%s"', [Ident] );
  end
  else Error( 'Identifier expected.' );
end;

procedure TInterpreter.CompileContinue;
begin
  ExecutionList.AddContinue;
  NextToken; // skip "continue" keyword
end;

procedure TInterpreter.CompileEvaluateBlock;
var
  DefText: string;
  Token: Word;
begin
  DefText := 'Print(' + TakePureText( Token, idScriptEnd ) + ')';
  {
  DefText := 'Print( ';
  Token := NextToken;
  while Token <> idSemicolon do
  begin
    if (Token = idEOF) or (Token = idScriptEnd) then Break;
    if Token = idString then
      DefText := DefText + GetQuotedStr(Parser.TokenString)
    else DefText := DefText + ' ' + Parser.TokenString;
    Token := NextToken;
  end;
  }
  if DefText <> '' then
    ExecutionList.AddCall('Print', DefText)
  else Error('Expression expected.');
end;

procedure TInterpreter.CompileFile(const FileName: string);
var
  Stream: TMemoryStream;
  Token: Word;
begin
  Stream := TMemoryStream.Create;
  try
    FScriptName := FileName;
    Stream.LoadFromFile(FileName);
    FParser := TCustomParser.Create(Stream);
    try
      while True do begin
        Token := NextToken;
        case Token of
          idEOF: Break;
        {
          idScriptBegin:;
          idScriptEnd:;
          idNullHtml:;
          idComment:;
          idSemicolon:; // Duzeltmeyi Unutma!
          idHtml: ExecutionList.AddHtml(Parser.TokenString);
          }
        else
          CompileBlock;
        end;
      end;
    finally
      Parser.Free;
    end;
  finally
    Stream.Free;
  end;
end;

procedure TInterpreter.CompileForBlock;
var
  Ident, Expr1, Expr2: string;
  Token: Word;
  WhileLine, EndLine: Integer;
  Item: TExecutionItem;

  { CompileBeginEndBlock }
  procedure CompileBeginEndBlock;
  begin
    Token := NextToken;
    while Token <> idEnd do
    begin
      if Token = idEOF then Break;
      CompileBlock;
      Token := NextToken;
    end;
    Token := NextToken; // Skip "end"
  end;

  { ScanBreakContinue }
  procedure ScanBreakContinue(const LoopLine, EndLine: Integer);
  var
    Index: Integer;
    Item: TExecutionItem;
  begin
    for Index := LoopLine to EndLine do
    begin
      Item := ExecutionList.Items[Index];
      if Item.Code = ecBreak then
      begin
        Item.Code := ecJump;
        Item.JumpCode := EndLine + 1;
      end
      else if Item.Code = ecContinue then
      begin
        Item.Code := ecJump;
        Item.JumpCode := LoopLine;
      end;
    end;
  end;

begin

  Token := NextToken;

  if Token <> idIdent then
    Error( 'Identifier expected.' );

  Ident := Parser.TokenString;
  if not ScriptVariantList.IsDefined(Ident) then
    ErrorFmt( 'Unknown identifier "%s".', [Ident] );

  Token := NextToken;
  if Token <> idAssignment then
    Error( '":=" expected.' );

  Expr1 := TakePureText( Token, idTo );
  // Make "To" error control!!!
  ExecutionList.AddAssign( Ident, Expr1 );

  Expr2 := TakePureText( Token, idDo );
  // Make "Do" error control!!!
  Token := NextToken; // Skip "do"

  WhileLine := ExecutionList.AddWhile( Ident + '<=' + Expr2 );
  
  if Token = idBegin then
    CompileBeginEndBlock
  else if Token = idSemiColon then
    Error( 'No Expression [For]' )
  else CompileBlock;

  ExecutionList.AddAssign( Ident, Ident + ' + 1' );
  //EndLine := ExecutionList.AddWend;
  EndLine := ExecutionList.AddJump;
  Item := ExecutionList.Items[WhileLine];
  Item.JumpCode := EndLine + 1;
  Item := ExecutionList.Items[EndLine];
  Item.JumpCode := WhileLine;

  ScanBreakContinue( WhileLine, EndLine );

end;

procedure TInterpreter.CompileFunctionBlock(const Ident: string);
var
  DefText: string;
  Token: Word;
begin

  if not ScriptFunctionList.IsDefined(Ident) then
    ErrorFmt( 'Unknown identifier: "%s"', [Ident] );

  DefText := Ident + TakePureText( Token, idSemicolon );
  if DefText <> '' then
    ExecutionList.AddCall(Ident, DefText)
  else Error('Expression expected.');
  
end;

procedure TInterpreter.CompileHtmlBlock;
begin
  ExecutionList.AddHtml(Parser.TokenString);
end;

procedure TInterpreter.CompileIdent;
var
  Ident: string;
begin
  Ident := Parser.TokenString;
  if ScriptVariantList.IsDefined(Ident) then begin
    CompileAssignBlock(Ident);
  end
  else if ScriptFunctionList.IsDefined(Ident) then begin
    CompileFunctionBlock(Ident);
  end
  else if ScriptConstantList.IsDefined(Ident) then begin
    ErrorFmt( 'Identifier "%s" is constant.', [Ident] );
  end
  else ErrorFmt( 'Unknown identifier: "%s"', [Ident] );
end;

procedure TInterpreter.CompileIfBlock;
var
  DefText: string;
  Token: Word;
  IfLine: Integer;

  { CompileBeginEndBlock }
  procedure CompileBeginEndBlock;
  begin
    Token := NextToken;
    while Token <> idEnd do
    begin
      if Token = idEOF then Break;
      CompileBlock;
      Token := NextToken;
    end;
    Token := NextToken; // Skip "end"
  end;

  { SetItemsLine }
  procedure SetItemsLine(const AIfLine, AJumpLine, AElseLine: Integer);
  var
    Item: TExecutionItem;
  begin
    Item := ExecutionList.Items[AJumpLine];
    Item.JumpCode := ExecutionList.Count;
    Item := ExecutionList.Items[AIfLine];
    Item.JumpCode := AElseLine;
  end;

  { CompileElseBlockForIfThen }
  procedure CompileElseBlockForIfThen;
  var
    JumpLine, ElseLine: Integer;
  begin
    ElseLine := -1;
    JumpLine := ExecutionList.AddJump;
    if Token = idSemiColon then
      ElseLine := ExecutionList.Count
    else if Token = idElse then
    begin
      ElseLine := ExecutionList.AddElse;
      Token := NextToken;
      if Token = idBegin then
        CompileBeginEndBlock
      else if Token = idSemiColon then
        //Error('Warning! No Expression (Else).') // Skip or Error.
      else
        CompileBlock;
    end
    else Error('";" expected.');
    SetItemsLine( IfLine, JumpLine, ElseLine );
  end;

begin
  DefText := TakePureText( Token, idThen );
  Token := NextToken;  // skip "then" keyword
  IfLine := ExecutionList.AddIf(DefText);
  if Token = idBegin then // with begin-end
    CompileBeginEndBlock
  else if Token = idElse then // Hemen else geliyor.
    Token := Parser.Token
  else if Token = idSemiColon then
    //Error('Warning! No Expression.') // Skip or Error
  else begin // One line if-then
    CompileBlock;
    Token := Parser.Token;
  end;
  CompileElseBlockForIfThen;
  
end;

procedure TInterpreter.CompileIncludeBlock;
var
  DefText: string;
  Token: Word;
  Interpreter: TInterpreter;
begin
  Token := NextToken;
  if Token = idString then begin
    DefText := Parser.TokenString;
    Interpreter := TInterpreter.Create;
    try
      DefText := Application.MapPath(DefText);
      Interpreter.CompileFile(DefText);
    finally
      Interpreter.Free;
    end;
  end;
end;

procedure TInterpreter.CompileRepeatBlock;
var
  DefText: string;
  Token: Word;
  RepeatLine: Integer;
  UntilLine: Integer;
  Item: TExecutionItem;

  { ScanBreakContinue }
  procedure ScanBreakContinue(const LoopLine, EndLine: Integer);
  var
    Index: Integer;
    Item: TExecutionItem;
  begin
    for Index := LoopLine to EndLine do
    begin
      Item := ExecutionList.Items[Index];
      if Item.Code = ecBreak then
      begin
        Item.Code := ecJump;
        Item.JumpCode := EndLine + 1;
      end
      else if Item.Code = ecContinue then
      begin
        Item.Code := ecJump;
        Item.JumpCode := LoopLine;
      end;
    end;
  end;

begin

  RepeatLine := ExecutionList.AddRepeat;

  Token := NextToken; // Skip Repeat keyword

  while Token <> idUntil do
  begin
    if Token = idEOF then Break;
    CompileBlock;
    Token := NextToken;
  end;

  DefText := TakePureText( Token, idSemiColon );

  UntilLine := ExecutionList.AddUntil( DefText );
  Item := ExecutionList.Items[UntilLine];
  Item.JumpCode := RepeatLine;

  ScanBreakContinue( RepeatLine, UntilLine );

end;

procedure TInterpreter.CompileVarBlock;
var
  Ident, DefText: string;
  Token: Word;
begin
  repeat
    Token := NextToken;

    if Token = idEOF then Break;

    if Token = idIdent then
    begin
      Ident := Parser.TokenString;
      if IdentIsKeyword(Ident) then
        ErrorFmt( 'Ident is keyword: "%s"', [Ident] );
      if not ScriptVariantList.IsDefined(DefText) then
        ScriptVariantList.CreateVar(Ident, Unassigned, idUnknown)
      else ErrorFmt( 'Variable already defined: "%s"', [Ident] );
      Continue;
    end;

    if Token = idEqual then
    begin
      DefText := TakePureText( Token, idSemiColon );
      ExecutionList.AddAssign(Ident, DefText);
      if Token = idEOF then Break;
    end;

    if Token = idSemiColon then Break;

    if Token = idComma then Continue;

  until Token = idSemiColon;

  if Token <> idSemiColon then Error( '";" expected.' );
  
end;

procedure TInterpreter.CompileWhileBlock;
var
  DefText: string;
  Token: Word;
  WhileLine: Integer;
  EndLine: Integer;
  Item: TExecutionItem;

  { ScanBreakContinue }
  procedure ScanBreakContinue(const LoopLine, EndLine: Integer);
  var
    Index: Integer;
    Item: TExecutionItem;
  begin
    for Index := LoopLine to EndLine do
    begin
      Item := ExecutionList.Items[Index];
      if Item.Code = ecBreak then
      begin
        Item.Code := ecJump;
        Item.JumpCode := EndLine + 1;
      end
      else if Item.Code = ecContinue then
      begin
        Item.Code := ecJump;
        Item.JumpCode := LoopLine;
      end;
    end;
  end;

begin

  DefText := TakePureText( Token, idDo );

  // Make "Do" error control!!!

  Token := NextToken;

  WhileLine := ExecutionList.AddWhile(DefText);

  if Token = idBegin then
  begin
    Token := NextToken;
    while Token <> idEnd do
    begin
      if Token = idEOF then Break;
      CompileBlock;
      Token := NextToken;
    end;
    NextToken; // For ";"
  end
  else if Token = idSemiColon then
    Error('Warning! Sonsuz dongu.')
  else CompileBlock;

  Item := ExecutionList.Items[WhileLine];
  EndLine := ExecutionList.AddJump;
  Item.JumpCode := EndLine + 1;
  Item := ExecutionList.Items[EndLine];
  Item.JumpCode := WhileLine;

  ScanBreakContinue( WhileLine, EndLine );

end;

constructor TInterpreter.Create;
begin
  inherited Create;
end;

destructor TInterpreter.Destroy;
begin
  inherited Destroy;
end;

procedure TInterpreter.Error(const Ident: string);
begin
  ErrorStr(Ident);
end;

procedure TInterpreter.ErrorFmt(const Ident: string; Args: array of const);
begin
  ErrorStr(Format(Ident, Args));
end;

procedure TInterpreter.ErrorStr(const Message: string);
begin
  raise EInterpreterError.CreateFmt( SInterpreterError, [ScriptName,
    Parser.Line, '<b>' + Message + '</b>'] );
end;

function TInterpreter.NextToken: Word;
begin
  repeat
    Result := Parser.NextToken;
  until Result <> idComment;
end;

procedure TInterpreter.Run;
var
  Index: Integer;
  Item: TExecutionItem;
begin
  Index := 0;
  while True do begin
    if (Index < 0) or (Index > ExecutionList.Count - 1) then Break;
    Item := TExecutionItem(ExecutionList.Items[Index]);
    {$IFDEF DEBUG}
      WriteLn(Format('%d -> Code: %d, %s, Jump: %d',
        [Index, Item.Code, Item.Text, Item.JumpCode]));
    {$ENDIF}
    case Item.Code of
      ecHtml: Application.WriteString(Item.Text);
      ecAssign: RunAssign(Item);
      ecCall: RunCall(Item);
      ecWhile:
        begin
          RunWhile(Item, Index);
          Continue;
        end;
      ecRepeat:
        begin
          RunRepeat(Item, Index);
          Continue;
        end;
      ecUntil:
        begin
          RunUntil(Item, Index);
          Continue;
        end;
      ecJump:
        begin
          RunJump(Item, Index);
          Continue;
        end;
      ecIf:
        begin
          RunIf(Item, Index);
          Continue;
        end;
      ecElse:
        begin
          RunElse(Item, Index);
          Continue;
        end;
      ecEndIf:
        begin
          RunEndIf(Item, Index);
          Continue;
        end;
      ecBreak:
        begin
          RunBreak(Item, Index);
          Continue;
        end;
      ecContinue:
        begin
          RunContinue(Item, Index);
          Continue;
        end;
    end;
    Inc(Index);
  end;
end;

procedure TInterpreter.RunAssign(Item: TExecutionItem);
var
  EvaluateExpr: TEvaluateExpr;
  Stream: TStringStream;
  Value: Variant;
  DefVariant: TScriptVariant;
begin
  Stream := TStringStream.Create('<% ' + Item.Text + ' %>' );
  try
    EvaluateExpr := TEvaluateExpr.Create(Self, Stream);
    try
      Value := EvaluateExpr.Evaluate;
      DefVariant := ScriptVariantList.GetVariant(Item.Ident);
      DefVariant.Value := Value;
      DefVariant.ObjectValue := EvaluateExpr.ObjectResult;
      DefVariant.VarType := EvaluateExpr.ResultType;
    finally
      EvaluateExpr.Free;
    end;
  finally
    Stream.Free;
  end;
end;

procedure TInterpreter.RunBreak(Item: TExecutionItem; var Index: Integer);
begin
  Index := Index + 1;
end;

procedure TInterpreter.RunCall(Item: TExecutionItem);
var
  EvaluateExpr: TEvaluateExpr;
  Stream: TStringStream;
begin
  Stream := TStringStream.Create('<% ' + Item.Text + ' %>' );
  try
    EvaluateExpr := TEvaluateExpr.Create(Self, Stream);
    try
      EvaluateExpr.Evaluate;
    finally
      EvaluateExpr.Free;
    end;
  finally
    Stream.Free;
  end;
end;

procedure TInterpreter.RunContinue(Item: TExecutionItem;
  var Index: Integer);
begin
  Index := Index + 1;
end;

procedure TInterpreter.RunElse(Item: TExecutionItem; var Index: Integer);
begin
  Index := Index + 1;
end;

procedure TInterpreter.RunEndIf(Item: TExecutionItem; var Index: Integer);
begin
  Index := Index + 1;
end;

procedure TInterpreter.RunIf(Item: TExecutionItem; var Index: Integer);
var
  EvaluateExpr: TEvaluateExpr;
  Stream: TStringStream;
  Value: Variant;
begin
  Stream := TStringStream.Create('<% ' + Item.Text + ' %>' );
  try
    EvaluateExpr := TEvaluateExpr.Create(Self, Stream);
    try
      Value := EvaluateExpr.Evaluate;
      if Value = False then
        Index := Item.JumpCode
      else Index := Index + 1;
    finally
      EvaluateExpr.Free;
    end;
  finally
    Stream.Free;
  end;
end;

procedure TInterpreter.RunJump(Item: TExecutionItem; var Index: Integer);
begin
  Index := Item.JumpCode;
end;

procedure TInterpreter.RunRepeat(Item: TExecutionItem; var Index: Integer);
begin
  Index := Index + 1;
end;

procedure TInterpreter.RunUntil(Item: TExecutionItem; var Index: Integer);
var
  EvaluateExpr: TEvaluateExpr;
  Stream: TStringStream;
  Value: Variant;
begin
  Stream := TStringStream.Create('<% ' + Item.Text + ' %>' );
  try
    EvaluateExpr := TEvaluateExpr.Create(Self, Stream);
    try
      Value := EvaluateExpr.Evaluate;
      if Value = False then
        Index := Item.JumpCode // Continue
      else Index := Index + 1; // Terminate
    finally
      EvaluateExpr.Free;
    end;
  finally
    Stream.Free;
  end;
end;

procedure TInterpreter.RunWhile(Item: TExecutionItem; var Index: Integer);
var
  EvaluateExpr: TEvaluateExpr;
  Stream: TStringStream;
  Value: Variant;
begin
  Stream := TStringStream.Create('<% ' + Item.Text + ' %>' );
  try
    EvaluateExpr := TEvaluateExpr.Create(Self, Stream);
    try
      Value := EvaluateExpr.Evaluate;
      if Value = False then
        Index := Item.JumpCode // Terminate
      else Index := Index + 1; // Continue
    finally
      EvaluateExpr.Free;
    end;
  finally
    Stream.Free;
  end;
end;

function TInterpreter.TakePureText(var CurrToken: Word;
  const ToToken: Word): string;
var
  SavePtr, EndPtr: PChar;
begin
  SavePtr := Parser.LastPtr;
  EndPtr := SavePtr;
  CurrToken := NextToken;
  while CurrToken <> ToToken do
  begin
    if (CurrToken = idEOF) or (CurrToken = idElse) then Break;
    EndPtr := Parser.LastPtr;
    CurrToken := NextToken;
  end;
  Result := Parser.GetPureText( SavePtr, EndPtr );
end;

initialization

finalization

end.
