unit ScriptParser;

interface

uses SysUtils, Classes;

const
  idUnknown         = $FFFF;
  idEOF             = 0;
  idSymbol          = 1;
  idString          = 2;
  idInteger         = 3;
  idFloat           = 4;
  idIdent           = 5;  // Same as idSymbol
  idComment         = 6;  // All comment types
  idHtml            = 7;
  idLessThan        = 8;  // <
  idScriptBegin     = 9;  // <%
  idScriptEnd       = 10; // %>
  idPercent         = 11; // %
  idSlash           = 12; // /
  idLBrace          = 13; // {
  idRBrace          = 14; // }
  idGreaterThan     = 15; // >
  idNullHtml        = 16; //
  idEqual           = 17; // =
  idNotEqual        = 18; // <>
  idAssignment      = 19; // :=
  idSemicolon       = 20; // ;
  idColon           = 21; // :
  idComma           = 22; // ,
  idPlus            = 23; // +
  idMinus           = 24; // -
  idMul             = 25; // *
  idDiv             = 26; // /
  idLessEqual       = 27; // <=
  idGreaterEqual    = 28; // >=
  idLanguage        = 29; // <%@
  idLParenthes      = 30; // (
  idRParenthes      = 31; // )
  idAt              = 32; // @
  idLeftBracket     = 33; // [
  idRightBracket    = 34; // ]
  idLeftBrace       = 35; // {
  idRightBrace      = 36; // }
  idTilde           = 37; // ~
  idExclamation     = 38; // !
  idDot             = 39; // .
  idTwoDots         = 40; // ..
  idEvaluate        = 41; // <%=

  idBitAnd          = 1000; // &
  idBitOr           = 1001; // |
  idBitNot          = 1002; // ~ idTilde;
  idBitXor          = 1003; // ^
  idShl             = 1004; // <<
  idShr             = 1005; // >>
  idMod             = idPercent; // %
  idAnd             = 2000; // &&
  idOr              = 2001; // ||
  idNot             = 2002; // !
  idXor             = 2003; // ^^

  idRegister        = 10000;
  idVariant         = 10001;
  idObject          = 10002;
  idBoolean         = 10003;

  idVar             = 3000;
  idConst           = 3001;
  idInclude         = 3002;
  idIf              = 3003;
  idThen            = 3004;
  idElse            = 3005;
  idEndIf           = 3006;
  idWhile           = 3008;
  idDo              = 3009;
  idWend            = 3010;
  idRepeat          = 3011;
  idUntil           = 3012;
  idEnd             = 3013;
  idElseIf          = 3014;
  idBegin           = 3015;
  idUses            = 3016;
  idBreak           = 3017;
  idContinue        = 3018;
  idFor             = 3019;
  idTo              = 3020;

type
  TToken = (
    toUnknown,
    toEOF,
    toSymbol,
    toString,
    toInteger,
    toFloat,
    toIdent,
    toComment,
    toHtml,
    toLessThan,
    toScriptBegin,
    toScriptEnd,
    toPercent,
    toSlash,
    toLBrace,
    toRBrace,
    toGreaterThan,
    toNullHtml,
    toEqual,
    toNotEqual,
    toAssignment,
    toSemicolon,
    toColon,
    toComma,
    toPlus,
    toMinus,
    toMultiply,
    toDivision,
    toLessEqual,
    toGreaterEqual,
    toLanguage,
    toLParenthes,
    toRParenthes,
    toAt,
    toLeftBracket,
    toRightBracket,
    toLeftBrace,
    toRightBrace,
    toTilde,
    toExclamation,
    toDot,
    toTwoDots,
    toEvaluate
  );
  
  TTokens = set of TToken;

  TReservedWordToken = ( toVar, toConst, toUses, toIf, toThen, toElse, toBegin,
    toEnd, toWhile, toDo, toRepeat, toUntil, toFor, toTo, toBreak, toContinue,
    toAnd, toOr, toXor, toShl, toShr, toNot, toMod, toDiv );

  TReservedWordTokens = set of toVar..toDiv; //TReservedWordToken

  //TMyTokens = TTokens * TReservedWordTokens;

type
  ECustomParserError = class(Exception);

  TCustomParser = class(TObject)
  private
    FStream: TStream;
    FBuffer: PChar;
    FLastPtr: PChar;
    FTokenPtr: PChar;
    FToken: Word;
    FLine: LongInt;
    FHtml: Boolean;
    FString: string;
  public
    constructor Create(Stream: TStream);
    destructor Destroy; override;
    function NextToken: Word;
    function TokenString: string;
    function TokenFloat: Extended;
    function TokenInt: Longint;
    procedure SkipBlanks;
    procedure SkipHtmlCode;
    function SourcePos: LongInt;
    function NextChar: Char;
    procedure Error(const Ident: string);
    procedure ErrorFmt(const Ident: string; Args: array of const);
    procedure ErrorStr(const Message: string);
    function GetPureText(StartPtr, EndPtr: PChar): string;
    property Token: Word read FToken;
    property Line: LongInt read FLine;
    property LastPtr: PChar read FLastPtr;
  end;

{ GetQuotedStr }
function GetQuotedStr(const S: string): string;

{ VarTypeToId }
function VarTypeToId(VarType: Integer): Word;

{ IdentIsKeyword }
function IdentIsKeyword(const Ident: string): Boolean;

implementation

uses Consts;

const
  SUnterminatedStringError = 'Unterminated string constant';
  SParseError = 'Parse Error: %s, (%d)';

type
  TKeywordRec = record
    Key: string;
    Ident: Word;
  end;

const
  MaxKeywords = 28;
  Keywords: array[0..MaxKeywords-1] of TKeywordRec =
    ((Key: 'not'; Ident: idNot),
     (Key: 'mod'; Ident: idMod),
     (Key: 'and'; Ident: idAnd),
     (Key: 'shl'; Ident: idShl),
     (Key: 'shr'; Ident: idShr),
     (Key: 'or';  Ident: idOr),
     (Key: 'xor'; Ident: idXor),
     (Key: 'div'; Ident: idDiv),

     (Key: 'var';       Ident: idVar),
     (Key: 'const';     Ident: idConst),
     (Key: 'include';   Ident: idInclude),
     (Key: 'if';        Ident: idIf),
     (Key: 'then';      Ident: idThen),
     (Key: 'else';      Ident: idElse),
     (Key: 'endif';     Ident: idEndIf),
     (Key: 'while';     Ident: idWhile),
     (Key: 'do';        Ident: idDo),
     (Key: 'wend';      Ident: idWend),
     (Key: 'repeat';    Ident: idRepeat),
     (Key: 'until';     Ident: idUntil),
     (Key: 'end';       Ident: idEnd),
     (Key: 'elseif';    Ident: idElseIf),
     (Key: 'begin';     Ident: idBegin),
     (Key: 'uses';      Ident: idUses),
     (Key: 'break';     Ident: idBreak),
     (Key: 'continue';  Ident: idContinue),
     (Key: 'for';       Ident: idFor),
     (Key: 'to';        Ident: idTo)
    );

{ GetQuotedStr }
function GetQuotedStr(const S: string): string;
var
  I: Integer;
begin
  Result := S;
  for I := Length(Result) downto 1 do
    if Result[I] = '"' then Insert('"', Result, I);
  Result := '"' + Result + '"';
end;

{ VarTypeToId }
function VarTypeToId(VarType: Integer): Word;
begin
  case VarType of
    varByte,
    varInteger,
    varSmallInt: Result := idInteger;
    varSingle,
    varDouble: Result := idFloat;
    varString: Result := idString;
  else Result := idUnknown;
  end;
end;

{ IdentIsKeyword }
function IdentIsKeyword(const Ident: string): Boolean;
var
  Index: Integer;
begin
  for Index := 0 to MaxKeywords - 1 do
  begin
    Result := SameText(Keywords[Index].Key, Ident);
    //CompareText(Keywords[Index].Key, Ident) = 0;
    if Result then Break;
  end;
end;

{ GetKeywordIdent }
function GetKeywordIdent(TokenPtr, LastPtr: PChar): Word;
var
  Key: string;
  Index: Integer;
begin
  Result := idIdent;
  SetString(Key, TokenPtr, LastPtr-TokenPtr);
  for Index := 0 to MaxKeywords-1 do
  begin
    if CompareText(Keywords[Index].Key, Key) = 0 then
    begin
      Result := Keywords[Index].Ident;
      Break;
    end;
  end;
end;

constructor TCustomParser.Create(Stream: TStream);
begin
  FStream := Stream;
  FBuffer := AllocMem(FStream.Size + 1);
  FStream.Seek(0, soFromBeginning);
  FStream.ReadBuffer(FBuffer^, FStream.Size);
  FStream.Seek(0, soFromBeginning);
  FLastPtr := FBuffer;
  FTokenPtr := FBuffer;
  FLine := 1;
  FHtml := True;
end;

destructor TCustomParser.Destroy;
begin
  FreeMem(FBuffer, FStream.Size + 1);
  inherited Destroy;
end;

function TCustomParser.NextToken: Word;
var
  I, J: Integer;
  P, S: PChar;
begin

  if FHtml then
  begin
    FTokenPtr := FLastPtr;
    SkipHtmlCode;
    Result := idHtml;
    if FTokenPtr = FLastPtr then Result := idNullHtml;
    FToken := Result;
    Exit;
  end;

  SkipBlanks;
  P := FLastPtr;
  FTokenPtr := P;

  case P^ of
    #0 : Result := idEOF;
    '@':
      begin
        Inc(P);
        Result := idAt;
      end;
    '.':
      begin
        Inc(P);
        Result := idDot;
        if P^ = '.' then
        begin
          Inc(P);
          Result := idTwoDots;
        end;
      end;
    '!':
      begin
        Inc(P);
        Result := idNot;
        if P^ = '=' then
        begin
          Inc(P);
          Result := idNotEqual;
        end;
      end;
    '|':
      begin
        Inc(P);
        Result := idBitOr;
        if P^ = '|' then
        begin
          Inc(P);
          Result := idOr;
        end;
      end;
    '&':
      begin
        Inc(P);
        Result := idBitAnd;
        if P^ = '&' then
        begin
          Inc(P);
          Result := idAnd;
        end;
      end;
    '~':
      begin
        Inc(P);
        Result := idBitNot;
      end;
    '^':
      begin
        Inc(P);
        Result := idBitXor;
        if P^ = '^' then
        begin
          Inc(P);
          Result := idXor;
        end;
      end;
    '[':
      begin
        Inc(P);
        Result := idLeftBracket;
      end;
    ']':
      begin
        Inc(P);
        Result := idRightBracket;
      end;
    '(':
      begin
        Inc(P);
        Result := idLParenthes;
      end;
    ')':
      begin
        Inc(P);
        Result := idRParenthes;
      end;
    '=':
      begin
        Inc(P);
        Result := idEqual;
      end;
    '+':
      begin
        Inc(P);
        Result := idPlus;
      end;
    '*':
      begin
        Inc(P);
        Result := idMul;
      end;
    ';':
      begin
        Inc(P);
        Result := idSemicolon;
      end;
    ',':
      begin
        Inc(P);
        Result := idComma;
      end;
    ':':
      begin
        Inc(P);
        Result := idColon;
        if P^ = '=' then
        begin
          Inc(P);
          Result := idAssignment;
        end;
      end;
    '>':
      begin
        Inc(P);
        Result := idGreaterThan;
        if P^ = '=' then
        begin
          Inc(P);
          Result := idGreaterEqual;
        end
        else if P^ = '>' then
        begin
          Inc(P);
          Result := idShr;
        end;
      end;
    '<':
      begin
        Result := idLessThan;
        Inc(P);
        case P^ of
          '%':
            begin
              Inc(P);
              FHtml := False;
              Result := idScriptBegin;
              if P^ = '@' then
              begin
                Inc(P);
                Result := idLanguage;
              end
              else if P^ = '=' then
              begin
                Inc(P);
                Result := idEvaluate;
              end;
            end;
          '>':
            begin
              Inc(P);
              Result := idNotEqual;
            end;
          '=':
            begin
              Inc(P);
              Result := idLessEqual;
            end;
          '<':
            begin
              Inc(P);
              Result := idShl;
            end;
        end;
      end;
    '%':
      begin
        Result := idPercent;
        Inc(P);
        if P^ = '>' then
        begin
          Inc(P);
          Result := idScriptEnd;
          FHtml := True;
        end;
      end;
    '{':
      begin
        Result := idLBrace;
        Inc(P);
        while True do
        begin
          case P^ of
            #0: Break;
            #10: Inc(FLine);
            '}':
              begin
                Result := idComment;
                Inc(P);
                Break;
              end;
          end;
          Inc(P);
        end;
      end;
    '/':
      begin
        Result := idSlash;
        Inc(P);
        case P^ of
          '/':
            begin
              Inc(P);
              while not (P^ in [#0, #13, #10]) do Inc(P);
              if P^ = #13 then Inc(P);
              Result := idComment;
            end;
          '*':
            begin
              Inc(P);
              while True do
              begin
                case P^ of
                   #0: Break;
                  #10: Inc(FLine);
                  '*':
                    begin
                      Inc(P);
                      if P^ = '/' then begin
                        Inc(P);
                        Break;
                      end;
                      Continue;
                    end;
                end;
                Inc(P);
              end;
              Result := idComment;
            end;
        end;
      end;
    '#', '''', '"':
      begin
        J := 0;
        S := P;
        while True do
          case P^ of
            '#':
              begin
                Inc(P);
                I := 0;
                while P^ in ['0'..'9'] do
                begin
                  I := I * 10 + (Ord(P^) - Ord('0'));
                  Inc(P);
                end;
                Inc(J);
              end;
            '''':
              begin
                Inc(P);
                while True do
                begin
                  case P^ of
                    #0, #10, #13:
                      Error(SUnterminatedStringError);
                    '''':
                      begin
                        Inc(P);
                        if P^ <> '''' then Break;
                      end;
                  end;
                  Inc(J);
                  Inc(P);
                end;
              end;
            '"':
              begin
                Inc(P);
                while True do
                begin
                  case P^ of
                    #0, #10, #13:
                      Error(SUnterminatedStringError);
                    '"':
                      begin
                        Inc(P);
                        if P^ <> '"' then Break;
                      end;
                  end;
                  Inc(J);
                  Inc(P);
                end;
              end;
          else
            Break;
          end;
        P := S;
        SetLength(FString, J);
        J := 1;
        while True do
          case P^ of
            '#':
              begin
                Inc(P);
                I := 0;
                while P^ in ['0'..'9'] do
                begin
                  I := I * 10 + (Ord(P^) - Ord('0'));
                  Inc(P);
                end;
                FString[J] := Char(I);
                Inc(J);
              end;
            '''':
              begin
                Inc(P);
                while True do
                begin
                  case P^ of
                    #0, #10, #13:
                      Error(SUnterminatedStringError);
                    '''':
                      begin
                        Inc(P);
                        if P^ <> '''' then Break;
                      end;
                  end;
                  FString[J] := P^;
                  Inc(J);
                  Inc(P);
                end;
              end;
            '"':
              begin
                Inc(P);
                while True do
                begin
                  case P^ of
                    #0, #10, #13:
                      Error(SUnterminatedStringError);
                    '"':
                      begin
                        Inc(P);
                        if P^ <> '"' then Break;
                      end;
                  end;
                  FString[J] := P^;
                  Inc(J);
                  Inc(P);
                end;
              end;
          else
            Break;
          end;
        Result := idString;
      end;
    'A'..'Z', 'a'..'z', '_':
      begin
        Inc(P);
        while P^ in ['A'..'Z', 'a'..'z', '0'..'9', '_'] do Inc(P);
        Result := GetKeywordIdent(FTokenPtr, P);
      end;
    '$':
      begin
        Inc(P);
        while P^ in ['0'..'9', 'A'..'F', 'a'..'f'] do Inc(P);
        Result := idInteger;
      end;
    '-':
      begin
        Result := idMinus;
        Inc(P);
      end;
    '0'..'9':
      begin
        while P^ in ['0'..'9'] do Inc(P);
        Result := idInteger;
        while P^ in ['0'..'9', '.'] do
        begin
          Inc(P);
          Result := idFloat;
        end;
        if P^ in ['e', 'E'] then
        begin
          while P^ in ['0'..'9', 'e', 'E', '+', '-'] do
          begin
            Inc(P);
            Result := idFloat;
          end;
        end;
      end;
  else
    Result := idUnknown;
    if P^ <> #0 then Inc(P);
  end;
  FLastPtr := P;
  FToken := Result;
end;

function TCustomParser.TokenString: string;
begin
  if FToken = idString then Result := FString
  else SetString(Result, FTokenPtr, FLastPtr - FTokenPtr);
end;

function TCustomParser.TokenFloat: Extended;
begin
  Result := StrToFloat(TokenString);
end;

function TCustomParser.TokenInt: Longint;
begin
  Result := StrToInt(TokenString);
end;

procedure TCustomParser.SkipBlanks;
begin
  while True do
  begin
    case FLastPtr^ of
      #0, #33..#255: Exit;
      #10: Inc(FLine);
    end;
    Inc(FLastPtr);
  end;
end;

procedure TCustomParser.SkipHtmlCode;
begin
  while True do
  begin
    case FLastPtr^ of
      #0:
        begin
          FHtml := False;
          Exit;
        end;
      #10: Inc(FLine);
      '<':
        begin
          if NextChar = '%' then
          begin
            FHtml := False;
            Exit;
          end;
        end;
    end;
    Inc(FLastPtr);
  end;
end;

function TCustomParser.SourcePos: LongInt;
begin
  Result := FLastPtr - FBuffer;
end;

function TCustomParser.NextChar: Char;
var
  P: PChar;
begin
  P := FLastPtr;
  Inc(P);
  Result := P^;
end;

procedure TCustomParser.Error(const Ident: string);
begin
  ErrorStr(Ident);
end;

procedure TCustomParser.ErrorFmt(const Ident: string; Args: array of const);
begin
  ErrorStr(Format(Ident, Args));
end;

procedure TCustomParser.ErrorStr(const Message: string);
begin
  raise ECustomParserError.CreateFmt(SParseError, [Message, FLine]);
end;

function TCustomParser.GetPureText(StartPtr, EndPtr: PChar): string;
begin
  SetString(Result, StartPtr, EndPtr - StartPtr);
end;

end.
