unit TreeList;

interface

uses SysUtils, Classes, Variants;

const
  ntInteger   = 1;
  ntFloat     = 2;
  ntString    = 3;
  ntPlus      = 4;
  ntMinus     = 5;
  ntMul       = 6;
  ntDiv       = 7;
  ntParenthes = 8;
  ntIdent     = 9;

type

  TNodeList = class;

  TNodeItem = class(TObject)
  private
    FNext: TNodeItem;
    FFirstChild: TNodeList;
    FValue: Variant;
    FObjectValue: TObject;
    FNodeType: Integer;
    FOwner: TNodeList;
  protected
    function GetHasChildren: Boolean;
    function GetLast: TNodeList;
    function GetPrev: TNodeItem;
  public
    constructor Create(AOwner: TNodeList);
    destructor Destroy; override;
    function AppendChild: TNodeList;
    procedure DeleteChild(Child: TNodeList);
    property Next: TNodeItem read FNext write FNext;
    property Prev: TNodeItem read GetPrev;
    property Value: Variant read FValue write FValue;
    property ObjectValue: TObject read FObjectValue write FObjectValue;
    property FirstChild: TNodeList read FFirstChild write FFirstChild;
    property HasChildren: Boolean read GetHasChildren;
    property Owner: TNodeList read FOwner;
    property NodeType: Integer read FNodeType write FNodeType;
    property LastChild: TNodeList read GetLast;
  end;

  TNodeList = class(TObject)
  private
    FParent: TNodeItem;
    FFirst: TNodeItem;
    FNext: TNodeList;
  protected
    function GetLast: TNodeItem;
    function GetPrev: TNodeList;
  public
    constructor Create(AParent: TNodeItem);
    destructor Destroy; override;
    function Append(AValue: Variant): TNodeItem;
    function GetCount: Integer;
    procedure Delete(Item: TNodeItem);
    property Parent: TNodeItem read FParent;
    property First: TNodeItem read FFirst write FFirst;
    property Last: TNodeItem read GetLast;
    property Next: TNodeList read FNext write FNext;
    property Prev: TNodeList read GetPrev;
  end;

implementation

{ TNodeList }

function TNodeList.Append(AValue: Variant): TNodeItem;
var
  P: TNodeItem;
begin
  Result := TNodeItem.Create(Self);
  Result.Value := AValue;
  if First <> nil then
  begin
    P := Last;
    P.Next := Result;
    {
    Result.Next := First;
    First := Result;
    }
  end
  else
    First := Result;
end;

constructor TNodeList.Create(AParent: TNodeItem);
begin
  inherited Create;
  FParent := AParent;
end;

procedure TNodeList.Delete(Item: TNodeItem);
var
  P: TNodeItem;
begin
  if Assigned(Item) then
  begin
    if Item = First then
    begin
      //First := Item.Next;
      First := First.Next;
      //First.Next := nil;
    end
    else
    if Item = Last then
    begin
      P := Item.Prev;
      P.Next := nil;
    end
    else begin
      P := Item.Prev;
      P.Next := Item.Next;
    end;
    FreeAndNil(Item);
  end;
end;

destructor TNodeList.Destroy;
var
  Item, P: TNodeItem;
begin
  Item := First;
  while Item <> nil do
  begin
    P := Item.Next;
    FreeAndNil(Item);
    Item := P;
  end;
  inherited Destroy;
end;

function TNodeList.GetCount: Integer;
var
  P: TNodeItem;
begin
  P := First;
  Result := 0;
  while P <> nil do
  begin
    Inc(Result);
    P := P.Next;
  end;
end;

function TNodeList.GetLast: TNodeItem;
var
  P: TNodeItem;
begin
  P := First;
  Result := P;
  while P <> nil do begin
    Result := P;
    P := P.Next;
  end;
end;

function TNodeList.GetPrev: TNodeList;
var
  P: TNodeList;
begin
  Result := nil;
  if Assigned(Parent) then
  begin
    P := Parent.FirstChild;
    while P <> nil do
    begin
      if P.Next = Self then
      begin
        Result := P;
        Break;
      end;
      P := P.Next;
    end;
  end;
end;

{ TNodeItem }

function TNodeItem.AppendChild: TNodeList;
var
  P: TNodeList;
begin
  Result := TNodeList.Create(Self);
  if FirstChild <> nil then
  begin
    P := LastChild;
    P.Next := Result;
  end
  else
    FFirstChild := Result;
end;

constructor TNodeItem.Create(AOwner: TNodeList);
begin
  inherited Create;
  FOwner := AOwner;
end;

procedure TNodeItem.DeleteChild(Child: TNodeList);
var
  P: TNodeList;
begin
  if Assigned(Child) then
  begin
    if Child = FirstChild then
    begin
      FirstChild := Child.Next;
    end
    else if Child = LastChild then
    begin
      P := Child.Prev;
      P.Next := nil;
    end
    else begin
      P := Child.Prev;
      P.Next := Child.Next;
    end;
    FreeAndNil(Child);
  end;
end;

destructor TNodeItem.Destroy;
var
  Item, P: TNodeList;
begin
  Item := FirstChild;
  while Item <> nil do begin
    P := Item.Next;
    FreeAndNil(Item);
    Item := P;
  end;
  inherited Destroy;
end;

function TNodeItem.GetHasChildren: Boolean;
begin
  Result := FirstChild <> nil;
end;

function TNodeItem.GetLast: TNodeList;
var
  P: TNodeList;
begin
  P := FirstChild;
  Result := P;
  while P <> nil do begin
    Result := P;
    P := P.Next;
  end;
end;

function TNodeItem.GetPrev: TNodeItem;
var
  P: TNodeItem;
begin
  Result := nil;
  if Assigned(Owner) then
  begin
    P := Owner.First;
    while P <> nil do
    begin
      if P.Next = Self then
      begin
        Result := P;
        Break;
      end;
      P := P.Next;
    end;
  end;
end;

end.
